# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import abc
import six
import string
import collections
import mimeparse
import weakref
from six.moves import http_cookies


class Headers(collections.Mapping):
    """Describes a mapping abstraction over request headers.
    """

    @staticmethod
    def _Header(sequence, name):
        """Returns the passed header as a tuple.

        Implements a facade so that the response headers can override
        this to provide a mutable sequence header.
        """
        return tuple(sequence._headers.get(name, '').split(','))

    class _Sequence(dict):
        """
        Provides an implementation of a dictionary that retreives its
        values as sequences.
        """

        def __init__(self, headers):
            self._headers = headers

        def __missing__(self, name):
            self[name] = value = self.headers._Header(self, name)
            return value

    def __init__(self):
        #! Internal store of the multi-valued headers as lists.
        self._sequence = self._Sequence(self)

    @staticmethod
    def normalize(name):
        """Normalizes the case of the passed name to be Http-Header-Case."""
        return str(string.capwords(name, '-'))

    @abc.abstractmethod
    def __getitem__(self, name):
        """Retrieves a header with the passed name.

        @param[in] name
            The case-insensitive name of the header to retrieve.
        """

    @abc.abstractmethod
    def __len__(self):
        """Retrieves the number of headers in the request."""

    @abc.abstractmethod
    def __iter__(self):
        """Returns an iterable for all headers in the request."""

    @abc.abstractmethod
    def __contains__(self, name):
        """Tests if the passed header exists in the request."""

    def index(self, name, value):
        """
        Return the index in the list of the first item whose value is x in
        the values of the named header.
        """
        return self._sequence[name].index(value)

    def count(self, name, value):
        """
        Return the number of times a value appears in the list of the values
        of the named header.
        """
        return self._sequence[name].count(value)

    def getlist(self, name):
        """Retrieves the passed header as a tuple of its values."""
        return self._sequence[name]


class Request(six.Iterator):
    """Describes the RESTful request abstraction.
    """

    #! Dictionary-like interface to access headers; this should be
    #! set by the dervied class to an instance of a derived Headers class.
    headers = None

    def __init__(self, path, method, asynchronous, *args, **kwargs):
        #! The captured path of the request, after the mount point.
        #! Example: GET /api/poll/23 => '/23'
        self.path = path

        #! True if we're asynchronous.
        self.asynchronous = asynchronous

        # Determine the actual HTTP method; apply the override header.
        override = self.headers.get('X-Http-Method-Override')
        if override:
            # Passed method was overriden; store override.
            self.method = override.upper()

        else:
            # Passed method is the actual method.
            self.method = method.upper()

        #! Cookie jar full of python morsel objects that represent the
        #! cookies that were sent with the request.
        text = self.get('Cookie')
        self.cookies = http_cookies.SimpleCookie()
        if text:
            self.cookies.load(str(text))

        #! A reference to the bound resource; this is set in the resource
        #! view method after traversal.
        self._resource = None

    def bind(self, resource):
        """Binds this to the passed resource object.

        This is used so that the request and response classes can access
        metadata and configuration on the resource handling this request
        so helper methods on the request and response like `serialize` work
        in full knowledge of configuration supplied to the resource.
        """
        self._resource = weakref.proxy(resource)

    @property
    def resource(self):
        return self._resource

    @property
    def protocol(self):
        """Retrieves the upper-cased version of the protocol (eg. HTTP)."""
        raise NotImplementedError()

    @property
    def host(self):
        """Retrieves the hostname, normally from the `Host` header."""
        return self.headers.get('Host') or '127.0.0.1'

    @property
    def mount_point(self):
        """Retrieves the mount point portion of the path of this request."""
        raise NotImplementedError()

    @property
    def query(self):
        """Retrieves the text after the first ? in the path."""
        raise NotImplementedError()

    @property
    def uri(self):
        """Returns the complete URI of the request."""
        raise NotImplementedError()

    @property
    def encoding(self):
        """
        The name of the encoding used to decode the stream’s bytes
        into strings, and to encode strings into bytes.

        Reads the charset value from the `Content-Type` header, if available;
        else, returns nothing.
        """
        # Get the `Content-Type` header, if available.
        content_type = self.headers.get('Content-Type')
        if content_type:
            # Parse out the primary type and parameters from the media type.
            ptype, _, params = mimeparse.parse_mime_type(content_type)

            # Return the specified charset or the default depending on the
            # primary type.
            default = 'utf-8' if ptype == 'application' else 'iso-8859-1'
            return params.get('charset', default)

    def _read(self):
        """Read and return the request data.

        @note Connectors should override this method.
        """
        return None

    def read(self, deserialize=False, format=None):
        """Read and return the request data.

        @param[in] deserialize
            True to deserialize the resultant text using a determiend format
            or the passed format.

        @param[in] format
            A specific format to deserialize in; if provided, no detection is
            done. If not provided, the content-type header is looked at to
            determine an appropriate deserializer.
        """

        if deserialize:
            data, _ = self.deserialize(format=format)
            return data

        content = self._read()

        if not content:
            return ''

        if type(content) is six.binary_type:
            content = content.decode(self.encoding)

        return content

    def deserialize(self, format=None):
        """Deserializes the request body using a determined deserializer.

        @param[in] format
            A specific format to deserialize in; if provided, no detection is
            done. If not provided, the content-type header is looked at to
            determine an appropriate deserializer.

        @returns
            A tuple of the deserialized data and an instance of the
            deserializer used.
        """
        return self._resource.deserialize(self, format=format)

    def __len__(self):
        """Returns the length of the request body, if known."""
        length = self.headers.get('Content-Length')
        return int(length) if length else 0

    def __getitem__(self, name):
        """Retrieves a header with the passed name."""
        return self.headers[name]

    def get(self, name, default=None):
        """Retrieves a header with the passed name."""
        return self.headers.get(name, default)

    def getlist(self, name):
        """
        Retrieves a the multi-valued list of the header with
        the passed name.
        """
        return self.headers.getlist(name)

    def __contains__(self, name):
        """Tests if the passed header exists in the request."""
        return name in self.headers

    def keys(self):
        """Return a new view of the header names."""
        return self.headers.keys()

    def values(self):
        """Return a new view of the header values."""
        return self.headers.values()

    def items(self):
        """Return a new view of the headers."""
        return self.headers.items()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import os
import wsgi_intercept
from wsgi_intercept.httplib2_intercept import install

# Setup the environment variables.
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.connectors.django.settings'

from .models import Poll

__all__ = [
    'Poll'
]


def http_setup(connectors, host, port, callback):
    # Setup the environment variables.
    os.environ['DJANGO_SETTINGS_MODULE'] = (
        'tests.connectors.django.settings')

    # Invoke the callback if we got one.
    if callback:
        callback()

    # Install the WSGI interception layer.
    install()

    # Import and grab the django application.
    from django.core.wsgi import get_wsgi_application

    # Enable the WSGI interception layer.
    wsgi_intercept.add_wsgi_intercept(host, port, get_wsgi_application)


def http_teardown(host, port):
    # Remove the WSGI interception layer.
    wsgi_intercept.remove_wsgi_intercept(host, port)


def model_setup():
    # Setup the environment variables.
    os.environ['DJANGO_SETTINGS_MODULE'] = (
        'tests.connectors.django.settings')

    # Initialize the database and create all models.
    from django.db import connections, DEFAULT_DB_ALIAS
    connection = connections[DEFAULT_DB_ALIAS]
    connection.creation.create_test_db(verbosity=0)

    # Load the data fixture.
    from django.core import management
    data = os.path.join(os.path.dirname(__file__), '..', 'data.json')
    management.call_command('loaddata', data, verbosity=0, interactive=0)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .base import ModelResource as BaseModelResource
from .meta import ModelResourceBase

__all__ = [
    'ModelResource'
]


class ModelResource(six.with_metaclass(ModelResourceBase, BaseModelResource)):
    """Implements the RESTful resource protocol for model resources.
    """



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from ._version import __version__, __version_info__  # noqa
from .decorators import route, resource, asynchronous
from .helpers import use

__all__ = [
    'route',
    'resource',
    'asynchronous',
    'use'
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


class Serializer(object):

    #! Applicable media types for this serializer.
    media_types = ()

    def __init__(self, request=None, response=None):
        #! The request and response objects to use.
        self.request = request
        self.response = response

    def can_serialize(self, data=None):
        """Tests this serializer to see if it can serialize."""
        try:
            # Attet to serialize the object.
            self.serialize(data)

            # The serialization process is assumed to have succeed.
            return True

        except ValueError:
            # The object was of an unsupported type.
            return False

    def serialize(self, data=None):
        """
        Transforms the object into an acceptable format for transmission.

        @throws ValueError
            To indicate this serializer does not support the encoding of the
            specified object.
        """
        if data is not None and self.response is not None:
            # Set the content type.
            self.response['Content-Type'] = self.media_types[0]

            # Write the encoded and prepared data to the response.
            self.response.write(data)

        # Return the serialized data.
        # This has normally been transformed by a base class.
        return data



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet.exceptions import ImproperlyConfigured
from ..managed import options


class ModelResourceOptions(options.ManagedResourceOptions):

    def __init__(self, meta, name, data, bases):
        # Initalize base resource options.
        super(ModelResourceOptions, self).__init__(meta, name, data, bases)

        #! Reference to the declarative model defined by
        #! the Object Relational Mapper (ORM).
        self.model = meta.get('model')
        if self.model is None and not self.abstract:
            raise ImproperlyConfigured(
                'Model resources must be bound to a model.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from django.db import models


class Poll(models.Model):

    question = models.CharField(max_length=1024)

    available = models.NullBooleanField()

    votes = models.IntegerField(null=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from .decorators import classproperty, boundmethod
from .functional import cons, compose
from .string import dasherize
from .package import import_module

__all__ = [
    'classproperty',
    'boundmethod',
    'cons',
    'compose',
    'import_module',
    'dasherize',
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import armet
from armet import utils
from . import options


#! Map of connector class objects in their connector modules.
CONNECTORS = {
    'http': {
        'resource': ('{}.resources', 'Resource', 'resource.base',),
        'options': ('{}.resources', 'ResourceOptions',)
    },
    'model': {
        'resource': ('{}.resources', 'ModelResource', 'model.base',),
        'options': ('{}.resources', 'ModelResourceOptions',)
    }
}


#! Set of connector classes.
connector_set = set()


def apply_connectors(meta, name, bases, metadata, base_meta):

    # Apply the declared connectors. What this entails is inserting the
    # connector base classes at the correct location in the base class list.

    # The correct location would be directly before one of the base resource
    # classes.

    # What is not implemented (and will not be) is replacing existing
    # connectors. You may have an inheritance chain of as many "abstract"
    # resources as warranted but as soon as a connector is supplied on a
    # non-abstract resource then that connector is "locked-in" so-to-speak.

    # Iterate through the declared connectors.
    cmap = CONNECTORS
    new_bases = list(bases)
    offset = 0
    for key, ref in six.iteritems(meta.connectors):

        try:
            # Resolve an optional options class that may be provided by
            # the connector.
            options = utils.import_module(cmap[key]['options'][0].format(ref))
            options = getattr(options, cmap[key]['options'][1])
            options_instance = options(metadata, name, base_meta)
            meta.__dict__.update(**options_instance.__dict__)

        except (AttributeError, KeyError):
            pass

        try:
            # Resolve the connector itself.
            module = utils.import_module(cmap[key]['resource'][0].format(ref))
            class_ = getattr(module, cmap[key]['resource'][1], None)

            # Ensure the connector has not already been applied.
            applied = False
            for base in bases:
                if issubclass(base, class_):
                    applied = True

            if applied:
                continue

            # Find where to insert this connector.
            module_key = 'armet.resources.%s' % cmap[key]['resource'][2]
            module = utils.import_module(module_key)
            armet_resource = getattr(module, cmap[key]['resource'][1])
            for index, base in enumerate(bases):
                if issubclass(base, armet_resource):
                    new_bases.insert(index + offset, class_)
                    offset += 1
                    break

        except (AttributeError, KeyError):
            pass

    return tuple(new_bases)

    # # Iterate through the available connectors.
    # cmap = CONNECTORS
    # for key, ref in six.iteritems(meta.connectors):
    #     options = utils.import_module(cmap[key]['options'][0].format(ref))
    #     if options:
    #         options = getattr(options, cmap[key]['options'][1], None)
    #         if options:
    #             # Available options to parse for this connector;
    #             # instantiate the options class and apply all
    #             # available options.
    #             options_instance = options(metadata, name, base_meta)
    #             meta.__dict__.update(**options_instance.__dict__)

    #     # Attempt to resolve the connector.
    #     module = utils.import_module(cmap[key]['resource'][0].format(ref))
    #     if module:
    #         class_ = getattr(module, cmap[key]['resource'][1], None)
    #         if class_:
    #             # Found a connector class for this connector.
    #             # Attempt to "apply" the connector.
    #             for base in bases:
    #                 if issubclass(base, class_):
    #                     break

    #             else:
    #                 # Mixin the class.
    #                 bases = (class_,) + bases



class ResourceBase(type):

    #! Options class to use to expand options.
    options = options.ResourceOptions

    #! Connectors to instantiate and mixin to the inheritance.
    connectors = ['http']

    @classmethod
    def _is_resource(cls, name, bases):
        if name == 'NewBase':
            # This is a six contrivance; not a real class.
            return False

        if name.startswith('armet.connector:'):
            # This is special mixed connector class; not something
            # to run the metaclass over.
            return False

        for base in bases:
            if base.__name__ == 'NewBase':
                # This is a six contrivance; move along.
                continue

            if isinstance(base, cls):
                # This is some sort of derived resource; good.
                return True

        # This is not derived at all from Resource (eg. is base).
        return False

    @classmethod
    def _gather_metadata(cls, metadata, bases):
        for base in bases:
            if isinstance(base, cls) and hasattr(base, 'Meta'):
                # Append metadata.
                metadata.append(getattr(base, 'Meta'))

                # Recurse.
                cls._gather_metadata(metadata, base.__bases__)

    def __new__(cls, name, bases, attrs):
        if not cls._is_resource(name, bases):
            # This is not an actual resource.
            return super(ResourceBase, cls).__new__(cls, name, bases, attrs)

        # Gather the attributes of all options classes.
        # Start with the base configuration.
        metadata = armet.use().copy()
        values = lambda x: {n: getattr(x, n) for n in dir(x)}

        # Expand the options class with the gathered metadata.
        base_meta = []
        cls._gather_metadata(base_meta, bases)

        # Apply the configuration from each class in the chain.
        for meta in base_meta:
            metadata.update(**values(meta))

        cur_meta = {}
        if attrs.get('Meta'):
            # Apply the configuration from the current class.
            cur_meta = values(attrs['Meta'])
            metadata.update(**cur_meta)

        # Gather and construct the options object.
        meta = attrs['meta'] = cls.options(metadata, name, cur_meta, base_meta)

        # Apply the found connectors.
        bases = apply_connectors(meta, name, bases, metadata, base_meta)

        # Construct the class object.
        self = super(ResourceBase, cls).__new__(cls, name, bases, attrs)

        # Generate a serializer map that maps media ranges to serializer
        # names.
        self._serializer_map = smap = {}
        for key in self.meta.allowed_serializers:
            serializer = self.meta.serializers[key]
            for media_type in serializer.media_types:
                smap[media_type] = key

        # Generate a deserializer map that maps media ranges to deserializer
        # names.
        self._deserializer_map = dmap = {}
        for key in self.meta.allowed_deserializers:
            deserializer = self.meta.deserializers[key]
            for media_type in deserializer.media_types:
                dmap[media_type] = key

        # Filter the available connectors according to the
        # metaclass restriction set.
        for key in list(meta.connectors.keys()):
            if key not in cls.connectors:
                del meta.connectors[key]

        # Return the constructed instance.
        return self



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .base import ManagedResource as BaseManagedResource
from .meta import ManagedResourceBase

__all__ = [
    'ManagedResource'
]


class ManagedResource(
        six.with_metaclass(ManagedResourceBase, BaseManagedResource)):
    """Implements the RESTful resource protocol for managed resources.
    """



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division

#! List of available HTTP/1.1 connectors.
http = ('bottle', 'flask', 'django',)

#! List of available ORM connectors.
model = ('django', 'sqlalchemy')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import json
from armet import http
from .base import BaseResourceTest
from pytest import mark
from armet.resources import ModelResource


@mark.bench('self.client.request', iterations=1000)
class TestResourceGet(BaseResourceTest):

    def test_list(self):
        response, content = self.client.request('/api/poll/')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 100
        assert data[0]['question'] == 'Are you an innie or an outie?'
        assert (data[-1]['question'] ==
                'What one question would you add to this survey?')

    def test_get_not_found(self, connectors):
        response, _ = self.client.get('/api/poll/101/')

        assert response.status == http.client.NOT_FOUND

    def test_single(self, connectors):
        response, content = self.client.request('/api/poll/1/')
        data = json.loads(content.decode('utf-8'))

        assert response.status == http.client.OK
        assert isinstance(data, dict)
        assert data['question'] == 'Are you an innie or an outie?'

        response, content = self.client.request('/api/poll/100/')
        data = json.loads(content.decode('utf-8'))

        assert response.status == http.client.OK
        assert isinstance(data, dict)
        assert (data['question'] ==
                'What one question would you add to this survey?')


@mark.bench('self.client.request', iterations=1000)
class TestResourceQuery(BaseResourceTest):

    def test_param_eq_one(self, connectors):
        response, content = self.client.request('/api/poll/?id=1')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 1
        assert data[0]['question'] == 'Are you an innie or an outie?'

    def test_param_eq_same_and(self, connectors):
        response, content = self.client.request('/api/poll/?id=1&id=2')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 0

    def test_param_eq_same_or(self, connectors):
        response, content = self.client.request('/api/poll/?id=1;id=2')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 2
        assert data[0]['question'] == 'Are you an innie or an outie?'

    def test_param_eq_same_or_simple(self, connectors):
        response, content = self.client.request('/api/poll/?id=1,2')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 2
        assert data[0]['question'] == 'Are you an innie or an outie?'

    def test_param_gt_one(self, connectors):
        response, content = self.client.request('/api/poll/?id>99')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 1

        question = data[0]['question']
        assert question == 'What one question would you add to this survey?'

    def test_param_available_true(self, connectors):
        response, content = self.client.request('/api/poll/?available=true')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 4

    def test_param_available_false(self, connectors):
        response, content = self.client.request('/api/poll/?available=false')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert isinstance(data, list)
        assert len(data) == 4


@mark.bench('self.client.request', iterations=1000)
class TestResourceTraversal(BaseResourceTest):

    def test_item_attribute(self):
        response, content = self.client.request('/api/poll/1/question/')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert data[0] == 'Are you an innie or an outie?'

    def test_item_attribute_not_found(self):
        response, content = self.client.request('/api/poll/1/blah/')

        assert response.status == http.client.NOT_FOUND

    def test_item_attribute_not_found_path(self):
        response, content = self.client.request('/api/poll/1/question/blah/')

        assert response.status == http.client.NOT_FOUND

    def test_list_attribute(self):
        response, content = self.client.request('/api/poll/question/')

        assert response.status == http.client.OK

        data = json.loads(content.decode('utf-8'))

        assert data[0] == 'Are you an innie or an outie?'
        assert data[-1] == 'What one question would you add to this survey?'

    def test_list_attribute_not_found(self):
        response, content = self.client.request('/api/poll/blah/')

        assert response.status == http.client.NOT_FOUND

    def test_list_attribute_not_found_path(self):
        response, content = self.client.request('/api/poll/question/blah/')

        assert response.status == http.client.NOT_FOUND


@mark.bench('self.client.request', iterations=1000)
class TestResourcePagination(BaseResourceTest):

    def test_get_range(self):
        response, content = self.client.request(
            path='/api/poll/',
            headers={
                'range': 'items=10-20'})

        assert response.status == http.client.PARTIAL_CONTENT

        data = json.loads(content.decode('utf-8'))

        assert len(data) == 11
        assert data[0]['question'] == 'Are you superstitious?'

    def test_get_single(self):
        response, content = self.client.request(
            path='/api/poll/',
            headers={
                'range': 'items=10'})

        assert response.status == http.client.PARTIAL_CONTENT

        data = json.loads(content.decode('utf-8'))

        assert len(data) == 1
        assert data[0]['question'] == 'Are you superstitious?'



# -*- coding: utf-8 -*-
"""
Implementation of a pagination interface using a combination of the HTTP/1.1
range header and set specifiers.
"""
from armet.http import exceptions, client

#! The range specifier to use.
RANGE_SPECIFIER = 'items'


def parse(specifiers):
    """
    Consumes set specifiers as text and forms a generator to retrieve
    the requested ranges.

    @param[in] specifiers
        Expected syntax is from the byte-range-specifier ABNF found in the
        [RFC 2616]; eg. 15-17,151,-16,26-278,15

    @returns
        Consecutive tuples that describe the requested range; eg. (1, 72) or
        (1, 1) [read as 1 to 72 or 1 to 1].
    """
    specifiers = "".join(specifiers.split())
    for specifier in specifiers.split(','):
        if len(specifier) == 0:
            raise ValueError("Range: Invalid syntax; missing specifier.")

        count = specifier.count('-')
        if (count and specifier[0] == '-') or not count:
            # Single specifier; return as a tuple to itself.
            yield int(specifier), int(specifier)
            continue

        specifier = list(map(int, specifier.split('-')))
        if len(specifier) == 2:
            # Range specifier; return as a tuple.
            if specifier[0] < 0 or specifier[1] < 0:
                # Negative indexing is not supported in range specifiers
                # as stated in the HTTP/1.1 Range header specification.
                raise ValueError(
                    "Range: Invalid syntax; negative indexing "
                    "not supported in a range specifier.")

            if specifier[1] < specifier[0]:
                # Range must be for at least one item.
                raise ValueError(
                    "Range: Invalid syntax; stop is less than start.")

            # Return them as a immutable tuple.
            yield tuple(specifier)
            continue

        # Something weird happened.
        raise ValueError("Range: Invalid syntax.")


def paginate(request, response, items):
    """Paginate an iterable during a request.

    Magically splicling an iterable in our supported ORMs allows LIMIT and
    OFFSET queries. We should probably delegate this to the ORM or something
    in the future.
    """
    # TODO: support dynamic rangewords and page lengths
    # TODO: support multi-part range requests

    # Get the header
    header = request.headers.get('Range')
    if not header:
        # No range header; move along.
        return items

    # do some validation
    prefix = RANGE_SPECIFIER + '='
    if not header.find(prefix) == 0:
        # This is not using a range specifier that we understand
        raise exceptions.RequestedRangeNotSatisfiable()

    else:
        # Chop the prefix off the header and parse it
        ranges = parse(header[len(prefix):])

    ranges = list(ranges)
    if len(ranges) > 1:
        raise exceptions.RequestedRangeNotSatisfiable(
            'Multiple ranges in a single request is not yet supported.')

    start, end = ranges[0]

    # Make sure the length is not higher than the total number allowed.
    max_length = request.resource.count(items)
    end = min(end, max_length)

    response.status = client.PARTIAL_CONTENT
    response.headers['Content-Range'] = '%d-%d/%d' % (start, end, max_length)
    response.headers['Accept-Ranges'] = RANGE_SPECIFIER

    # Splice and return the items.
    items = items[start - 1:end]
    return items



# -*- coding: utf-8 -*-
"""
Describes the authorization protocols and generalizations used to
authorize access to a resource endpoint.
"""
from __future__ import absolute_import, unicode_literals, division
from armet.exceptions import ImproperlyConfigured
from armet import http


class Authorization(object):
    """
    Establishes the protocol for authorization; by default, all users are
    authorized.
    """

    def is_accessible(self, user, method, resource):
        """
        Determines the accessibility to a resource endpoint for a particular
        method. An inaccessible resource is indistinguishable from a
        non-existant resource.

        @param[in] user
            The user in question that is being checked.

        @param[in] method
            The method in question that is being performed (eg. 'GET').

        @param[in] resource
            The resource instance that is being authorized.

        @returns
            Returns true if the user can access the resource for
            the passed operation; otherwise, false.
        """
        return True

    def inaccessible(self):
        """Informs the client that the resource is inaccessible."""
        raise http.exceptions.Forbidden()

    def is_authorized(self, user, operation, resource, item):
        """Determines authroization to a specific resource object.

        @param[in] user
            The user in question that is being checked.

        @param[in] operation
            The operation in question that is being performed (eg. 'read').

        @param[in] resource
            The resource instance that is being authorized.

        @param[in] item
            The specific instance of an object returned by a `read` from
            the `resource`.

        @returns
            Returns true to indicate authorization or false to indicate
            otherwise.
        """
        return True

    def unauthorized(self):
        """Informs the client that it is not authrozied for the resource."""
        raise http.exceptions.Forbidden()

    def filter(self, user, operation, resource, iterable):
        """
        Filters an iterable to contain only the items for which the user
        is authorized to perform the operation on.

        @param[in] user
            The user in question that is being checked.

        @param[in] operation
            The operation in question that is being performed (eg. 'read').

        @param[in] resource
            The resource instance that is being authorized.

        @param[in] iterable
            The iterable of objects to be checked. This method is called
            from the model connector so the actual value of this parameter
            depends on the model connector (eg. it may be a django queryset).

        @returns
            Returns an iterable containing the remaining objects.
        """
        return iterable


class ManagedAuthorization(Authorization):
    """Extends the authorization protocol to apply to managed resources.

    Managed resources only check accessibility and authorization based on
    the operation (eg. 'read') and not the method (eg. 'GET').
    """

    def inaccessible(self):
        """Informs the client that the resource is inaccessible."""
        raise http.exceptions.NotFound()


try:
    import shield

except ImportError:
    shield = None


class ShieldAuthorization(ManagedAuthorization):
    """Implements managed authorization using the shield permissions framework.
    """

    def __init__(self, permissions=None):
        # Ensure we have access to the shield framework.
        if shield is None:
            raise ImproperlyConfigured(
                "'shield' is required to use 'ShieldAuthorization'")

        #! Permissions to check for each operation.
        self.permissions = permissions
        if self.permissions is None:
            self.permissions = {
                'read': ('read',),
                'create': ('create',),
                'update': ('update',),
                'destroy': ('destroy',)
            }

    def is_authorized(self, user, operation, resource, item):
        return shield.has(
            *self.permissions[operation], bearer=user, target=item)

    def filter(self, user, operation, resource, iterable):
        clause = shield.filter(
            *self.permissions[operation],
            bearer=user, target=resource.meta.model)

        return resource.filter(clause, iterable)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import unittest
from armet import resources
from pytest import mark


# @mark.bench('self.Resource.parse', iterations=10000)
@mark.bench('self.request', iterations=10000)
class ParseTestCase(unittest.TestCase):

    def setUp(self):
        super(ParseTestCase, self).setUp()

        # Construct a resource.
        class Resource(resources.ManagedResource):
            class Meta:
                abstract = True

        # Store the resource
        self.Resource = Resource

    def request(self, url):
        # Execute the parse method.
        _, arguments, _ = self.Resource.parse(url)

        # Assert that we got enough of them.
        assert 'slug' in arguments
        assert 'query' in arguments
        assert 'extensions' in arguments
        assert 'directives' in arguments
        assert 'path' in arguments

        # Return the arguments
        return arguments

    def assert_arguments(self, arguments, **kwargs):
        for argument in kwargs:
            value = arguments.pop(argument)
            assert value == kwargs[argument]

        for name, value in six.iteritems(arguments):
            test = [] if name.endswith('s') else None
            assert value == test

    def test_simple(self):
        arguments = self.request('/')

        self.assert_arguments(arguments)

    def test_slug(self):
        arguments = self.request('/123/')

        self.assert_arguments(arguments, slug='123')

    def test_slug_long(self):
        arguments = self.request('/sdf-sdg-sgh-sh234-bf/')

        self.assert_arguments(arguments, slug='sdf-sdg-sgh-sh234-bf')

    def test_query(self):
        arguments = self.request('(stuff=32)/')

        self.assert_arguments(arguments, query='stuff=32')

    def test_query_long(self):
        query = 'x=32&y=61;z=135,15&x=16;!(h=134&b=324)'
        arguments = self.request('({})/'.format(query))

        self.assert_arguments(arguments, query=query)

    def test_query_directives(self):
        query = 'x:asc=32&y:desc=61;z:-=135,15&x:descending=16;!(h=134&b=324)'
        arguments = self.request('({})/'.format(query))

        self.assert_arguments(arguments, query=query)

    def test_directive(self):
        arguments = self.request(':random/')

        self.assert_arguments(arguments, directives=['random'])

    def test_directives(self):
        directives = ['random', 'rand', 'other', 'weird', 'wan']
        url = ':{}/'.format(':'.join(directives))
        arguments = self.request(url)

        self.assert_arguments(arguments, directives=directives)

    def test_directives_query(self):
        directives = ['random', 'rand', 'other', 'weird', 'wan']
        url = ':{}(x=4&y=234&z=34)/'.format(':'.join(directives))
        arguments = self.request(url)

        self.assert_arguments(
            arguments, directives=directives, query='x=4&y=234&z=34')

    def test_path(self):
        path = 'this/is/a/path/to/somewhere'
        arguments = self.request('/124/{}/'.format(path))

        self.assert_arguments(arguments, slug='124', path=path)

    def test_query_path(self):
        query = 'x=34&t:desc=324&(x=234&x:asc=2134)'
        path = 'this/is/a(x=324)/path(y=3124&x=23)/to/somewhere'
        arguments = self.request('({})/555/{}/'.format(
            query, path))

        self.assert_arguments(arguments, slug='555', query=query, path=path)

    def test_path_query(self):
        path = 'this/is/a(x=324)/path(y=3124&x=23)/to/somewhere'
        arguments = self.request('/124/{}/'.format(path))

        self.assert_arguments(arguments, slug='124', path=path)

    def test_extension(self):
        ext = 'json'
        arguments = self.request('.{}/'.format(ext))

        self.assert_arguments(arguments, extensions=[ext])

    def test_extensions(self):
        exts = ['schema', 'json']
        arguments = self.request('.{}/'.format('.'.join(exts)))

        self.assert_arguments(arguments, extensions=exts)

    def test_extensions_slug(self):
        exts = ['schema', 'json']
        arguments = self.request('/234.{}/'.format('.'.join(exts)))

        self.assert_arguments(arguments, slug='234', extensions=exts)

    def test_extensions_path(self):
        exts = ['schema', 'json']
        url = '/234/from/this/to/that.{}/'.format('.'.join(exts))
        arguments = self.request(url)

        self.assert_arguments(
            arguments, slug='234', extensions=exts,
            path='from/this/to/that')

    def test_all(self):
        extensions = ['schema', 'json']
        directives = ['random', 'rand', 'okay', 'because']
        slug = '2342fsdg8920sd-235-f-g325-sg23-532'
        path = '/from/this(x=235)/to:random/that(because)/and/this'
        query = 'x=345&xv=32;!(x=34)'
        url = ':{}({})/{}/{}.{}/'.format(
            ':'.join(directives),
            query,
            slug,
            path,
            '.'.join(extensions))
        arguments = self.request(url)

        self.assert_arguments(
            arguments,
            slug=slug,
            extensions=extensions,
            directives=directives,
            path=path,
            query=query)

    def test_extension_path(self):
        url = '/23/from/this.hub/32/from.hub/32/'
        arguments = self.request(url)

        self.assert_arguments(
            arguments,
            slug='23',
            path='from/this.hub/32/from.hub/32')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import ujson as json
from collections import Iterable, Sequence, Mapping
from .base import Serializer
from armet import media_types


class JSONSerializer(Serializer):

    media_types = media_types.JSON

    def serialize(self, obj=None):
        # If we have nothing; serialize as an empty object.
        if obj is None:
            return '{}'

        # Ensure generators are evaluated.
        if (isinstance(obj, Iterable)
                and not isinstance(obj, Sequence)
                and not isinstance(obj, Mapping)):
            obj = list(obj)

        # Ensure it is atleast wrapped in an array.
        if isinstance(obj, six.string_types) or not isinstance(obj, Iterable):
            obj = [obj]

        # Serialize the resultant text.
        text = json.dumps(obj, ensure_ascii=False)

        # Return us to the base to enclose it inside of a response object.
        return super(JSONSerializer, self).serialize(text)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .attribute import Attribute
import decimal


class DecimalAttribute(Attribute):

    type = decimal.Decimal

    def prepare(self, value):
        if value is None:
            return None

        return six.text_type(float(value))

    def clean(self, value):
        if isinstance(value, decimal.Decimal):
            # Value is already an integer.
            return value

        if isinstance(value, six.string_types):
            # Strip the string of whitespace
            value = value.strip()

        if not value:
            # Value is nothing; return nothing.
            return None

        try:
            # Attempt to coerce whatever we have as an int.
            return decimal.Decimal(value)

        except ValueError:
            # Failed to do so.
            raise ValueError('Not a valid decimal value.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from six.moves import cStringIO as StringIO
import operator
import collections
from itertools import chain
from . import constants


#! Operation to combinator map.
COMBINATORS = {
    constants.LOGICAL_AND: operator.and_,
    constants.LOGICAL_OR: operator.or_}


#! Set of characters that begin an operator.
OPERATOR_BEGIN_CHARS = set(x[0] for _, x in constants.OPERATORS if x)
OPERATOR_BEGIN_CHARS.add(constants.NEGATION[1])

#! Dictionary of operator symbols to operators.
OPERATOR_SYMBOL_MAP = dict((v, k) for k, v in constants.OPERATORS if v)


#! List of operator keywords.
OPERATOR_KEYWORDS = set(k for k, _ in constants.OPERATORS)


class Query(collections.Sequence):
    """Represents a complete query expression.
    """

    def __init__(self, segments=None):
        #! The various query segments.
        self.segments = [] if segments is None else segments

    def __getitem__(self, index):
        return self.segments[index]

    def __len__(self):
        return len(self.segments)

    def __repr__(self):
        return str(self)

    def __str__(self):
        o = StringIO()

        o.write('(')

        for index, segment in enumerate(self):

            o.write(str(segment))

            if index + 1 < len(self):
                comb = '&' if segment.combinator == operator.and_ else '|'
                o.write(' {} '.format(comb))

        o.write(')')

        return o.getvalue()


def parse(text, encoding='utf8'):
    """Parse the querystring into a normalized form."""
    # Initialize the query object.
    query = Query()

    # Decode the text if we got bytes.
    if isinstance(text, six.binary_type):
        text = text.decode(encoding)

    # Iterate through the characters in the query string; one-by-one
    # in order to perform one-pass parsing.
    stream = StringIO()

    for character in text:

        # We want to stop reading the query and pass it off to someone
        # when we reach a logical or grouping operator.
        if character in (constants.LOGICAL_AND, constants.LOGICAL_OR):

            if not stream.tell():
                # There is no content in the stream; a logical operator
                # was found out of place.
                raise ValueError('Found `{}` out of place'.format(
                    character))

            # Parse the segment up till the combinator
            segment = parse_segment(stream.getvalue(), character)
            query.segments.append(segment)
            stream.truncate(0)
            stream.seek(0)

        else:
            # This isn't a special character, just roll with it.
            stream.write(character)

    # TODO: Throw some nonsense here if the query string ended with a
    # & or ;, because that makes no sense.

    if stream.tell():
        # Append the remainder of the query string.
        query.segments.append(parse_segment(stream.getvalue()))

    # Return the constructed query object.
    return query


class QuerySegment(object):
    """
    Represents a single query segment with a subject path (`x.a.g`),
    an operator (`=` or `<=`), optional directives (`sort`), and
    a set of values (`5,12,56`).
    """

    def __init__(self, **kwargs):
        #! Path to the attribute being tested (as a list of segments).
        self.path = kwargs.get('path', [])

        #! This is the operator that is being applied to the attribute path.
        self.operator = kwargs.get('operator', constants.OPERATOR_IEQUAL[0])

        #! Negation; if this operation has been negated.
        self.negated = kwargs.get('negated', False)

        #! Directives. Directives are a key-value way of specifying commands
        #! on an attribute path.
        self.directives = kwargs.get('directives', [])

        #! Values. Set of values that the attribute path is being checked
        #! against. Only one has to match.
        self.values = kwargs.get('values', [])

        #! The combinator that is used to combine this and the next
        #! query.
        self.combinator = kwargs.get('combinator', operator.and_)

    def __repr__(self):
        return str(self)

    def __str__(self):
        """
        Format this query segment in a human-readable representation
        intended for debugging.
        """

        o = StringIO()

        if self.negated:
            o.write('not ')

        o.write('.'.join(self.path))

        if self.values:
            o.write(' :%s ' % self.operator)

        o.write(' | '.join(map(repr, self.values)))
        return o.getvalue()


def _parse_operator(segment, iterator):
    """Parses the operator (eg. '==' or '<')."""
    stream = StringIO()
    for character in iterator:
        if character == constants.NEGATION[1]:
            if stream.tell():
                # Negation can only occur at the start of an operator.
                raise ValueError('Unexpected negation.')

            # We've been negated.
            segment.negated = not segment.negated
            continue

        if (stream.getvalue() + character not in OPERATOR_SYMBOL_MAP and
                stream.getvalue() + character not in OPERATOR_BEGIN_CHARS):
            # We're no longer an operator.
            break

        # Expand the operator
        stream.write(character)

    # Check for existance.
    text = stream.getvalue()
    if text not in OPERATOR_SYMBOL_MAP:
        # Doesn't exist because of a mis-placed negation in the middle
        # of the path.
        raise ValueError('Unexpected negation.')

    # Set the found operator.
    segment.operator = OPERATOR_SYMBOL_MAP[text]

    # Return the remaining characters.
    return chain(character, iterator)


def parse_segment(text, combinator=constants.LOGICAL_AND):
    # Initialize a query segment.
    segment = QuerySegment()

    # Construct an iterator over the segment text.
    iterator = iter(text)
    stream = StringIO()

    # Iterate through the characters in the segment; one-by-one
    # in order to perform one-pass parsing.
    for character in iterator:

        if (character == constants.NEGATION[1]
                and not stream.tell() and not segment.path):
            # We've been negated.
            segment.negated = not segment.negated
            continue

        if character in OPERATOR_BEGIN_CHARS:
            # Found an operator; pull out what we can.
            iterator = _parse_operator(segment, chain(character, iterator))

            # We're done here; go to the value parser
            break

        if character == constants.SEP_PATH:
            # A path separator, push the current stack into the path
            segment.path.append(stream.getvalue())
            stream.truncate(0)
            stream.seek(0)

            # Keep checking for more path segments.
            continue

        # Append the text to the stream
        stream.write(character)

    # Write any remaining information into the path.
    segment.path.append(stream.getvalue())

    # Attempt to normalize the path.
    try:
        # The keyword 'not' can be the last item which
        # negates this query.
        if segment.path[-1] == constants.NEGATION[0]:
            segment.negated = not segment.negated
            segment.path.pop(-1)

        # The last keyword can explicitly state the operation; in which
        # case the operator symbol **must** be `=`.
        if segment.path[-1] in OPERATOR_KEYWORDS:
            if segment.operator != constants.OPERATOR_IEQUAL[0]:
                raise ValueError(
                    'Explicit operations must use the `=` symbol.')

            segment.operator = segment.path.pop(-1)

        # Make sure we still have a path left.
        if not segment.path:
            raise IndexError()

    except IndexError:
        # Ran out of path items after removing operations and negation.
        raise ValueError('No path specified in {}'.format(text))

    # Values are not complicated (yet) so just slice and dice
    # until we get a list of possible values.
    segment.values = ''.join(iterator)
    if segment.values:
        segment.values = segment.values.split(constants.SEP_VALUE)

    # Set the combinator.
    segment.combinator = COMBINATORS[combinator]

    # Return the constructed query segment.
    return segment



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet import exceptions
from .attribute import Attribute


try:
    # TODO: List this somewhere as an optional dep.
    import pytz

except ImportError:
    # No timezone support.
    pytz = None


class TimezoneAttribute(Attribute):

    if pytz is not None:
        type = pytz.tzfile.DstTzInfo

    def __init__(self, *args, **kwargs):
        # Ensure we have support.
        if pytz is None:
            raise exceptions.ImproperlyConfigured(
                'Use of the timezone attribute requires pytz')

        # Continue on.
        super(TimezoneAttribute, self).__init__(*args, **kwargs)

    def prepare(self, value):
        return value.zone

    def clean(self, value):
        if value is None:
            # Value is nothing; return it.
            return value

        try:
            # Attempt to coerce the timezone.
            return pytz.timezone(value)

        except pytz.UnknownTimeZoneError:
            raise ValueError('Unknown time zone.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import operator
import six
from six.moves import map, reduce
from django.conf import urls
from django.db.models import Q
from django.views.decorators import csrf
from armet import utils
from . import http
from armet.query import parser, Query, QuerySegment, constants


class Resource(object):

    @classmethod
    @csrf.csrf_exempt
    def view(cls, django_request, *args, **kwargs):
        # Construct request and response wrappers.
        async = cls.meta.asynchronous
        path = kwargs.get('path') or ''
        request = http.Request(django_request, path=path, asynchronous=async)
        response = http.Response(asynchronous=async)

        # Defer the execution thread if we're running asynchronously.
        if async:
            # Defer the view to pass of control.
            import gevent
            gevent.spawn(super(Resource, cls).view, request, response)

            # Construct and return the generator response.
            response._handle.content = cls.stream(response, response)
            return response._handle

        # Pass control off to the resource handler.
        result = super(Resource, cls).view(request, response)

        # Configure the response and return it.
        response._handle.content = result
        return response._handle

    @utils.classproperty
    def urls(cls):
        """Builds the URL configuration for this resource."""
        return urls.patterns('', urls.url(
            r'^{}(?:$|(?P<path>[/:(.].*))'.format(cls.meta.name),
            cls.view,
            name='armet-api-{}'.format(cls.meta.name),
            kwargs={'resource': cls.meta.name}))


# Build an operator map to use for django.
OPERATOR_MAP = {
    constants.OPERATOR_EQUAL[0]: '__exact',
    constants.OPERATOR_IEQUAL[0]: '__iexact',
    constants.OPERATOR_LT[0]: '__lt',
    constants.OPERATOR_GT[0]: '__gt',
    constants.OPERATOR_LTE[0]: '__lte',
    constants.OPERATOR_GTE[0]: '__gte',
}


def build_clause(query, attributes):
    # Iterate through each query segment.
    clause = None
    last = None
    for seg in query.segments:
        # Get the attribute in question.
        attribute = attributes[seg.path[0]]

        # Replace the initial path segment with the expanded
        # attribute path.
        seg.path[0:1] = attribute.path.split('.')

        # Boolean's should use `exact` rather than `iexact`.
        if attribute.type is bool:
            op = '__exact'
        else:
            op = OPERATOR_MAP[seg.operator]

        # Build the path from the segment.
        path = '__'.join(seg.path) + op

        # Construct a Q-object from the segment.
        q = reduce(operator.or_,
                   map(lambda x: Q((path, x)),
                       map(attribute.try_clean, seg.values)))

        # Combine the segment with the last.
        clause = last.combinator(clause, q) if last is not None else q
        last = seg

    # Return the constructed clause.
    return clause


class ModelResource(object):

    def filter(self, clause, queryset):
        # Filter the queryset by the passed clause.
        return queryset.filter(clause).distinct()

    def count(self, queryset):
        # Return the count of the queryset.
        return len(queryset)

    def read(self):
        # Initialize the queryset to the model manager.
        queryset = self.meta.model.objects

        query = None
        if self.slug is not None:
            # This is an item-access (eg. GET /<name>/:slug); ignore the
            # query string and generate a query-object based on the slug.
            query = Query(segments=[QuerySegment(
                path=self.meta.slug.path.split('.'),
                operator=constants.OPERATOR_EQUAL[0],
                values=[self.slug])])

        elif self.request.query:
            # This is a list-access; use the query string and construct
            # a query object from it.
            query = parser.parse(self.request.query)

        # Determine if we need to filter the queryset in some way; and if so,
        # filter it.
        if query is not None:
            clause = build_clause(query, self.attributes)
            queryset = self.filter(clause, queryset)

        # Filter the queryset by asserting authorization.
        queryset = self.meta.authorization.filter(
            self.request.user, 'read', self, queryset)

        if self.slug is not None:
            # Attempt to return just the single result we should have.
            result = queryset.all()[:1]
            return result[0] if result else None

        # Return the entire queryset.
        return queryset.all()

    def create(self, data):
        # Instantiate a new target.
        target = self.meta.model()

        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            value = data.get(name)
            if value is not None:
                attribute.set(target, value)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'create', self, target):
            authz.unauthorized()

        # Save the target.
        target.save()

        # Return the target.
        return target

    def update(self, target, data):
        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            attribute.set(target, data.get(name))

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'update', self, target):
            authz.unauthorized()

        # Save the target.
        target.save()

    def destroy(self):
        # Grab the existing target.
        target = self.read()

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'destroy', self, target):
            authz.unauthorized()

        # Destroy the target.
        target.delete()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import re


def dasherize(value):
    """Dasherizes the passed value."""
    value = value.strip()
    value = re.sub(r'([A-Z])', r'-\1', value)
    value = re.sub(r'[-_\s]+', r'-', value)
    value = re.sub(r'^-', r'', value)
    value = value.lower()
    return value



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import operator
from functools import partial
from six.moves import map, reduce
from armet.exceptions import ImproperlyConfigured
from armet.query import parser, Query, QuerySegment, constants


class ModelResourceOptions(object):

    def __init__(self, meta, name, bases):
        #! SQLAlchemy session used to perform operations on the models.
        self.Session = meta.get('Session')
        if not self.Session:
            raise ImproperlyConfigured(
                'A session factory (via sessionmaker) is required by '
                'the SQLAlchemy model connector.')


def iequal_helper(x, y):
    # String values should use ILIKE queries.
    if isinstance(type(y), six.string_types):
        return x.ilike(y)
    else:
        return operator.eq(x, y)


# Build an operator map to use for sqlalchemy.
OPERATOR_MAP = {
    constants.OPERATOR_EQUAL[0]: operator.eq,
    constants.OPERATOR_IEQUAL[0]: iequal_helper,
    constants.OPERATOR_LT[0]: operator.lt,
    constants.OPERATOR_GT[0]: operator.gt,
    constants.OPERATOR_LTE[0]: operator.le,
    constants.OPERATOR_GTE[0]: operator.ge,
}


def build_segment(model, segment, attr, clean):
    # Get the associated column for the initial path.
    path = segment.path.pop(0)
    col = getattr(model, path)

    # Resolve the inner-most path segment.
    if segment.path:
        if col.impl.accepts_scalar_loader:
            return col.has(build_segment(
                col.property.mapper.class_, segment, attr, clean))

        else:
            return col.any(build_segment(
                col.property.mapper.class_, segment, attr, clean))

    # Determine the operator.
    op = OPERATOR_MAP[segment.operator]

    # Apply the operator to the values and return the expression
    return reduce(operator.or_,
                  map(partial(op, col),
                      map(lambda x: clean(attr.try_clean(x)), segment.values)))


def build_clause(resource, query, attributes, cleaners, model):
    # Iterate through each query segment.
    clause = None
    last = None
    for seg in query.segments:
        # Get the attribute in question.
        attribute = attributes[seg.path[0]]

        # Replace the initial path segment with the expanded
        # attribute path.
        seg.path[0:1] = attribute.path.split('.')

        # Construct the clause from the segment.
        clean = partial(cleaners[attribute.name], resource)
        q = build_segment(model, seg, attribute, clean)

        # Combine the segment with the last.
        clause = last.combinator(clause, q) if last is not None else q
        last = seg

    # Return the constructed clause.
    return clause


class ModelResource(object):
    """Specializes the RESTFul model resource protocol for SQLAlchemy.

    @note
        This is not what you derive from to create resources. Import
        ModelResource from `armet.resources` and derive from that.
    """

    def route(self, *args, **kwargs):
        # Establish a session.
        self.session = session = self.meta.Session()

        try:
            # Continue on with the cycle.
            result = super(ModelResource, self).route(*args, **kwargs)

            # Commit the session.
            session.commit()

            # Return the result.
            return result

        except:
            # Something occurred; rollback the session.
            session.rollback()

            # Re-raise the exception.
            raise

        finally:
            # Close the session.
            session.close()

    def filter(self, clause, queryset):
        # Filter the queryset by the passed clause.
        return queryset.filter(clause).distinct()

    def count(self, queryset):
        # Return the count of the queryset.
        return queryset.count()

    def read(self):
        # Initialize the query to the model.
        queryset = self.session.query(self.meta.model)

        query = None
        if self.slug is not None:
            # This is an item-access (eg. GET //:slug); ignore the
            # query string and generate a query-object based on the slug.
            query = Query(segments=[QuerySegment(
                path=self.meta.slug.path.split('.'),
                operator=constants.OPERATOR_EQUAL[0],
                values=[self.slug])])

        elif self.request.query:
            # This is a list-access; use the query string and construct
            # a query object from it.
            query = parser.parse(self.request.query)

        # Determine if we need to filter the queryset in some way; and if so,
        # filter it.
        clause = None
        if query is not None:
            clause = build_clause(
                self, query, self.attributes,
                self.cleaners, self.meta.model)

            queryset = self.filter(clause, queryset)

        # Filter the queryset by asserting authorization.
        queryset = self.meta.authorization.filter(
            self.request.user, 'read', self, queryset)

        # Return the queryset.
        return queryset if self.slug is None else queryset.first()

    def create(self, data):
        # Instantiate a new target.
        target = self.meta.model()

        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            value = data.get(name)
            if value is not None:
                attribute.set(target, value)

        # Add the target to the session.
        self.session.add(target)
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'create', self, target):
            authz.unauthorized()

        # Return the target.
        return target

    def update(self, target, data):
        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            attribute.set(target, data.get(name))

        # Flush the target and expire attributes.
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'update', self, target):
            authz.unauthorized()

    def destroy(self):
        # Grab the existing target.
        target = self.read()

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'destroy', self, target):
            authz.unauthorized()

        # Remove the object from the session.
        self.session.delete(target)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from ..managed import meta
from . import options


class ModelResourceBase(meta.ManagedResourceBase):

    options = options.ModelResourceOptions

    connectors = ['http', 'model']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


def use(**kwargs):
    """
    Updates the active resource configuration to the passed
    keyword arguments.

    Invoking this method without passing arguments will just return the
    active resource configuration.

    @returns
        The previous configuration.
    """
    config = dict(use.config)
    use.config.update(kwargs)
    return config

# Set the initial resource configuration.
use.config = {}



# -*- coding: utf-8 -*-
"""Normalizes access to the HTTP libraries.
"""
from __future__ import absolute_import, unicode_literals, division
from six.moves import http_client as client
from .request import Request
from .response import Response
from . import exceptions

__all__ = [
    'client',
    'Request',
    'Response',
    'exceptions'
]

try:
    # Attempt to get additional status codes (added in python 3.2)
    getattr(client, 'PERMANENT_REDIRECT')
    getattr(client, 'PRECONDITION_REQUIRED')
    getattr(client, 'TOO_MANY_REQUESTS')
    getattr(client, 'REQUEST_HEADER_FIELDS_TOO_LARGE')
    getattr(client, 'NETWORK_AUTHENTICATION_REQUIRED')

except AttributeError:
    # Don't have em; add them.
    client.PERMANENT_REDIRECT = 308
    client.PRECONDITION_REQUIRED = 428
    client.TOO_MANY_REQUESTS = 429
    client.REQUEST_HEADER_FIELDS_TOO_LARGE = 431
    client.NETWORK_AUTHENTICATION_REQUIRED = 511
